(* $Id: godi_conffile.ml 678 2007-02-25 17:16:54Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_query

let read_godi_conf ?(pref_user_error=false) () =
  let gc_filename = godi_conf() in
  let f = 
    try
      open_in_bin gc_filename
    with
	Sys_error msg when pref_user_error ->
	  user_error ("Reading godi.conf: " ^ msg)
  in
  let lines = ref [] in
  try
    while true do
      let line = input_line f in
      lines := line :: !lines
    done;
    assert false
  with
      End_of_file ->
	close_in f;
	List.rev !lines
    | error ->
	close_in f;
	raise error
;;


let write_godi_conf ?(pref_user_error=false) ?(destfile = godi_conf()) gc =
  let gc_filename = destfile in
  let f = 
    try
      open_out_bin gc_filename 
    with
	Sys_error msg when pref_user_error ->
	  user_error ("Writing godi.conf: " ^ msg)
  in
  try
    List.iter
      (fun line ->
	 output_string f line;
	 output_string f "\n";
      )
      gc;
    close_out f
  with
      error ->
	close_out f;
	raise error
;;


let assignment_re = Str.regexp "^\\([A-Za-z_0-9]+\\)[ \t]*=";;

let ins_point_re = Str.regexp "^## GODI_INSERTION_POINT";;

let ins_point_str = "## GODI_INSERTION_POINT (do not delete this line)";;

let endif_re = Str.regexp "^\\.endif";;


let enter_option name opt gc =
  let ok = ref false in

  (* First trial: Is there already an assignment? *)

  let gc' =
    List.map
      (fun line ->
	 if Str.string_match assignment_re line 0 then (
	   let lname = Str.matched_group 1 line in
	   if lname = name then (
	     ok := true;
	     name ^ "=" ^ opt
	   )
	   else line
	 )
	 else line
      )
      gc in

  if !ok then 
    gc'
  else (

    (* Second trial: Insert at the insertion point *)
    
    let gc' =
      List.flatten
	(List.map
	   (fun line ->
	      if Str.string_match ins_point_re line 0 then (
		ok := true;
		[ name ^ "=" ^ opt; line ]
	      )
	      else [line]
	   )
	   gc
	) in

    if !ok then
      gc'
    else (

      (* Third trial: Insert at end *)
	
      gc @ [ name ^ "=" ^ opt; ins_point_str ]

    )
  )
;;


let modify_godi_conf ?(pref_user_error = false) ?destfile pkg_list =
  let gc = ref (read_godi_conf ~pref_user_error ()) in
  let mflag = ref false in
  List.iter
    (fun pkg ->
       match pkg with
	   `Source_pkg p ->
	     let curr_opts = p # values_of_build_options in
	     let new_opts = p # values_of_user_build_options in
	     let both_opts = 
	       List.map
		 (fun ((n,a),(n',b)) ->
		    assert(n=n');
		    (n,a,b))
		 (List.combine curr_opts new_opts) in
	     let changed_opts =
	       List.filter
		 (fun (n,a,b) -> a <> b) both_opts in
	     List.iter
	       (fun (n,a,b) ->
		  gc := enter_option n b !gc
	       )
	       changed_opts;
	     mflag := !mflag || changed_opts <> [];
	 | _ ->
	     ()
    )
    pkg_list;
  write_godi_conf ~pref_user_error ?destfile !gc;
  !mflag
;;
